;; Generated package description from code-cells.el  -*- no-byte-compile: t -*-
(define-package "code-cells" "0.3.0.20231015.132845" "Lightweight notebooks with support for ipynb files" '((emacs "27.1")) :commit "8a1f0281f88239b956250a8c76bada8aa9debe2c" :authors '(("Augusto Stoffel" . "arstoffel@gmail.com")) :maintainer '("Augusto Stoffel" . "arstoffel@gmail.com") :keywords '("convenience" "outlines") :url "https://github.com/astoff/code-cells.el")
